/* zoom.c:
 *
 * zoom an image
 *
 * jim frost 10.11.89
 *
 * Copyright 1989 Jim Frost.  See included file "copyright.h" for complete
 * copyright information.
 */

#include "copyright.h"
#include "image.h"

int smooth_scaling = 1;

static unsigned int *buildIndex(width, zoom, rwidth)
     unsigned int  width;
     float         zoom;
     unsigned int *rwidth;
{ float         fzoom;
  unsigned int *index;
  unsigned int  a;

  if (!zoom) {
    fzoom= 100.0;
    *rwidth= width;
  }
  else {
    fzoom= (float)zoom / 100.0;
    *rwidth= fzoom * width;
  }
  index= (unsigned int *)lmalloc(sizeof(unsigned int) * *rwidth);
  for (a= 0; a < *rwidth; a++)
    if (zoom)
      *(index + a)= (float)a / fzoom;
    else
      *(index + a)= a;
  return(index);
}

static Pixel rgb8_to_true(r, g, b)
     byte r;
     byte g;
     byte b;
{
  return (Pixel)r << 16 | (Pixel)g << 8 | (Pixel)b;
}

/* simple bilinear resampling */
static void resize_image(oimage, image)
     Image *oimage;
     Image *image;
{
  double cx, cy, xoff, yoff;
  double xstep, ystep;
  int maxx, maxy;
  int x1, x2, y1, y2;
  int i, j;
  byte *line1, *line2;
  byte *destptr;
  byte *adest = NULL;
  unsigned int srclinelen;
  unsigned int pixlen;
  Pixel v[4];
  unsigned int r, g, b;

  maxx = oimage->width - 1;
  maxy = oimage->height - 1;
  pixlen= oimage->pixlen;
  srclinelen= oimage->width * pixlen;
  destptr = image->data;
  if (oimage->trans != -1 || oimage->alpha) {
    image->trans = -1;
    lfree(image->alpha);
    image->alpha = lmalloc(image->width * image->height);
    adest = image->alpha;
  }
  cy = 0;
  xstep = (double)oimage->width / (double)image->width;
  ystep = (double)oimage->height / (double)image->height;
  for (j = image->height; j > 0; j--) {
    y1 = cy;
    if (y1 >= maxy) {
      y1 = y2 = maxy;
      yoff = 0;
    } else {
      y2 = y1 + 1;
      yoff = cy - y1;
    }
    line1 = oimage->data + srclinelen * y1;
    line2 = oimage->data + srclinelen * y2;
    cx = 0;
    for (i = image->width; i > 0; i--) {
      byte a[4];
      memset(a, 0xff, 4);
      x1 = cx;
      if (x1 >= maxx) {
	x1 = x2 = maxx;
	xoff = 0;
      } else {
	x2 = x1 + 1;
	xoff = cx - x1;
      }
      v[0] = memToVal(line1 + pixlen * x1, pixlen);
      if (v[0] == oimage->trans) {
	v[0] = 0;
	a[0] = 0;
      }
      v[1] = memToVal(line1 + pixlen * x2, pixlen);
      if (v[1] == oimage->trans) {
	v[1] = 0;
	a[1] = 0;
      }
      v[2] = memToVal(line2 + pixlen * x1, pixlen);
      if (v[2] == oimage->trans) {
	v[2] = 0;
	a[2] = 0;
      }
      v[3] = memToVal(line2 + pixlen * x2, pixlen);
      if (v[3] == oimage->trans) {
	v[3] = 0;
	a[3] = 0;
      }
      r = (TRUE_RED(v[0]) * (1 - xoff) +
	   TRUE_RED(v[1]) * xoff) * (1 - yoff) +
	(TRUE_RED(v[2]) * (1 - xoff) +
	 TRUE_RED(v[3]) * xoff) * yoff;
      g = (TRUE_GREEN(v[0]) * (1 - xoff) +
	   TRUE_GREEN(v[1]) * xoff) * (1 - yoff) +
	(TRUE_GREEN(v[2]) * (1 - xoff) +
	 TRUE_GREEN(v[3]) * xoff) * yoff;
      b = (TRUE_BLUE(v[0]) * (1 - xoff) +
	   TRUE_BLUE(v[1]) * xoff) * (1 - yoff) +
	(TRUE_BLUE(v[2]) * (1 - xoff) +
	 TRUE_BLUE(v[3]) * xoff) * yoff;
      v[0] = rgb8_to_true(r, g, b);
      valToMem(v[0], destptr, pixlen);
      destptr += pixlen;

      if (adest) {
	if (oimage->alpha) {
	  byte *alpha1 = oimage->alpha + oimage->width * y1;
	  byte *alpha2 = oimage->alpha + oimage->width * y2;
	  a[0] = alpha1[x1];
	  a[1] = alpha1[x2];
	  a[2] = alpha2[x1];
	  a[3] = alpha2[x2];
	}

	*adest++ = (byte)((a[0] * (1 - xoff) + a[1] * xoff) * (1 - yoff) +
			  (a[2] * (1 - xoff) + a[3] * xoff) * yoff);
      }

      cx += xstep;
    }
    cy += ystep;
  }
}

/* reduce the given image in integer factor */
static Image *reduced_image(oimage, width, height)
     Image *oimage;
     unsigned int width;
     unsigned int height;
{
  Image *dest;
  byte *destptr;
  byte *adest = NULL;
  unsigned int srclinelen;
  unsigned int i, j;
  unsigned int factorx = oimage->width / width;
  unsigned int factory = oimage->height / height;
  unsigned int pixlen= oimage->pixlen;

  if (factorx < 1)
    factorx = 1;
  if (factory < 1)
    factory = 1;

  width = oimage->width / factorx;
  if (oimage->width % factorx)
    width++;
  height = oimage->height / factory;
  if (oimage->height % factory)
    height++;

  dest = newTrueImage(width, height);
  destptr = dest->data;
  if (oimage->trans) {
    dest->alpha = (unsigned char *)lmalloc(width * height);
    adest = dest->alpha;
  }

  srclinelen = oimage->width * oimage->pixlen;

  for (j = 0; j < oimage->height; j += factory) {
    for (i = 0; i < oimage->width; i += factorx) {
      unsigned int r, g, b, a, div, x, y;
      Pixel v;

      r = g = b = a = 0;
      div = 0;
      for (y = 0; y < factory && j + y < oimage->height; y++) {
	byte *src = oimage->data + srclinelen * (j + y);
	byte *asrc = NULL;
	if (oimage->alpha)
	  asrc = oimage->alpha + oimage->width * (j + y);
	for (x = 0; x < factorx && i + x < oimage->width; x++) {
	  v = memToVal(src + pixlen * (i + x), pixlen);
	  if (asrc) {
	    if (*asrc) {
	      r += TRUE_RED(v);
	      g += TRUE_GREEN(v);
	      g += TRUE_BLUE(v);
	    }
	    a += *asrc;
	    asrc++;
	  } else if (v != oimage->trans) {
	    r += TRUE_RED(v);
	    g += TRUE_GREEN(v);
	    b += TRUE_BLUE(v);
	    a += 0xff;
	  }
	  div++;
	}
      }
      r /= div;
      g /= div;
      b /= div;
      v = rgb8_to_true(r, g, b);
      valToMem(v, destptr, pixlen);
      destptr += pixlen;
      if (adest)
	*adest++ = a / div;
    }
  }
  return dest;
}

Image *zoom(oimage, xzoom, yzoom, verbose)
     Image        *oimage;
     float        xzoom, yzoom;
{ char          buf[BUFSIZ];
  Image        *image=NULL;
  unsigned int *xindex, *yindex;
  unsigned int  xwidth, ywidth;
  unsigned int  x, y, xsrc, ysrc;
  unsigned int  pixlen;
  unsigned int  srclinelen;
  unsigned int  destlinelen;
  byte         *srcline, *srcptr;
  byte         *destline, *destptr;
  byte          srcmask, destmask, bit;
  Pixel         value;

  goodImage(oimage, "zoom");

  if ((xzoom == 0.0) && (yzoom == 0.0)) /* stupid user */
    return(NULL);

  if (xzoom == 0.0) {
    if (verbose)
      fprintf(stderr, "  Zooming image Y axis by %0.2f%%...", yzoom);
    if (oimage->title)
      sprintf(buf, "%s (Y zoom %0.2f%%)", oimage->title, yzoom);
  }
  else if (yzoom == 0.0) {
    if (verbose)
      fprintf(stderr, "  Zooming image X axis by %0.2f%%...", xzoom);
    if (oimage->title)
      sprintf(buf, "%s (X zoom %0.2f%%)", oimage->title, xzoom);
  }
  else if (xzoom == yzoom) {
    if (verbose)
      fprintf(stderr, "  Zooming image by %0.2f%%...", xzoom);
    if (oimage->title)
      sprintf(buf, "%s (%d%% zoom)", oimage->title, xzoom);
  }
  else {
    if (verbose)
      fprintf(stderr, "  Zooming image X axis by %0.2f%% and Y axis by %0.2f%%...",
	     xzoom, yzoom);
    if (oimage->title)
      sprintf(buf, "%s (X zoom %0.2f%% Y zoom %0.2f%%)", oimage->title,
	      xzoom, yzoom);
  }
  if (verbose)
    fflush(stderr);

  xindex= buildIndex(oimage->width, xzoom, &xwidth);
  yindex= buildIndex(oimage->height, yzoom, &ywidth);

  switch (oimage->type) {
  case IBITMAP:
    image= newBitImage(xwidth, ywidth);
    for (x= 0; x < oimage->rgb.used; x++) {
      *(image->rgb.red + x)= *(oimage->rgb.red + x);
      *(image->rgb.green + x)= *(oimage->rgb.green + x);
      *(image->rgb.blue + x)= *(oimage->rgb.blue + x);
    }
    image->rgb.used= oimage->rgb.used;
    destline= image->data;
    destlinelen= (xwidth / 8) + (xwidth % 8 ? 1 : 0);
    srcline= oimage->data;
    srclinelen= (oimage->width / 8) + (oimage->width % 8 ? 1 : 0);
    for (y= 0, ysrc= *(yindex + y); y < ywidth; y++) {
      while (ysrc != *(yindex + y)) {
	ysrc++;
	srcline += srclinelen;
      }
      srcptr= srcline;
      destptr= destline;
      srcmask= 0x80;
      destmask= 0x80;
      bit= srcmask & *srcptr;
      for (x= 0, xsrc= *(xindex + x); x < xwidth; x++) {
	if (xsrc != *(xindex + x)) {
	  do {
	    xsrc++;
	    if (!(srcmask >>= 1)) {
	      srcmask= 0x80;
	      srcptr++;
	    }
	  } while (xsrc != *(xindex + x));
	  bit= srcmask & *srcptr;
	}
	if (bit)
	  *destptr |= destmask;
	if (!(destmask >>= 1)) {
	  destmask= 0x80;
	  destptr++;
	}
      }
      destline += destlinelen;
    }
    break;

  case IRGB:
    image= newRGBImage(xwidth, ywidth, oimage->depth);
    for (x= 0; x < oimage->rgb.used; x++) {
      *(image->rgb.red + x)= *(oimage->rgb.red + x);
      *(image->rgb.green + x)= *(oimage->rgb.green + x);
      *(image->rgb.blue + x)= *(oimage->rgb.blue + x);
    }
    image->rgb.used= oimage->rgb.used;
    /* FALLTHRU */

  case ITRUE:
    if (!RGBP(oimage)) {
      Image *tmpimage = NULL;
      if (smooth_scaling &&
	  (oimage->width / xwidth > 1 || oimage->height / ywidth > 1)) {
	tmpimage = reduced_image(oimage, xwidth, ywidth);
	if (tmpimage->width == xwidth && tmpimage->height == ywidth) {
	  image = tmpimage;
	  break;
	}
      }
      image= newTrueImage(xwidth, ywidth);
      if (smooth_scaling) {
	resize_image(tmpimage ? tmpimage : oimage, image);
	if (tmpimage)
	  freeImage(tmpimage);
	break;
      }
    }

    pixlen= oimage->pixlen;
    destptr= image->data;
    srcline= oimage->data;
    srclinelen= oimage->width * pixlen;
    for (y= 0, ysrc= *(yindex + y); y < ywidth; y++) {
      while (ysrc != *(yindex + y)) {
	ysrc++;
	srcline += srclinelen;
      }

      srcptr= srcline;
      value= memToVal(srcptr, pixlen);
      for (x= 0, xsrc= *(xindex + x); x < xwidth; x++) {
	if (xsrc != *(xindex + x)) {
	  do {
	    xsrc++;
	    srcptr += image->pixlen;
	  } while (xsrc != *(xindex + x));
	  value= memToVal(srcptr, pixlen);
	}
	valToMem(value, destptr, pixlen);
	destptr += pixlen;
      }
    }
    break;
  }

  if (image) {
      image->title= dupString(buf);
      if (!image->alpha)
	image->trans = oimage->trans;
  }
  lfree((byte *)xindex);
  lfree((byte *)yindex);
  if (verbose)
    fprintf(stderr, "done\n");
  return(image);
}
